"""GbibParse
  Parse a BibTeX file once, to increase standard compliance. Use "GbibClean" for a more rigorous
  clean-up.

Usage:
  GbibParse <input> <output>

Arguments:
  input     Input BibTeX-file.
  output    Output file, or path (the "input" file-name is copied).

Options:
      --version   Show version.
  -h, --help      Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""

# ==================================================================================================

import os
import re
import sys
import bibtexparser
import docopt

from .. import __version__

# ==================================== RAISE COMMAND LINE ERROR ====================================

def Error(msg,exit_code=1):

  print(msg)

  sys.exit(exit_code)

# ========================================== MAIN PROGRAM ==========================================

def main():

  # --------------------------------- parse command line arguments ---------------------------------

  # parse command-line options/arguments
  args = docopt.docopt(__doc__, version=__version__)

  # change keys to simplify implementation:
  # - remove leading "-" and "--" from options
  args = {re.sub(r'([\-]{1,2})(.*)',r'\2',key): args[key] for key in args}
  # - change "-" to "_" to facilitate direct use in print format
  args = {key.replace('-','_'): args[key] for key in args}
  # - remove "<...>"
  args = {re.sub(r'(<)(.*)(>)',r'\2',key): args[key] for key in args}

  # --------------------------------------- check arguments ----------------------------------------

  # check that the BibTeX file exists
  if not os.path.isfile(args['input']):
    Error('"{input:s}" does not exist'.format(**args))

  # if "output" is an existing directory; convert to file with same name as the input file
  if os.path.isdir(args['output']):
    args['output'] = os.path.join(args['output'], os.path.split(args['input'])[-1])

  # ---------------------------------------- parse bib-file ----------------------------------------

  # read
  bib = bibtexparser.load(open(args['input'],'r'), parser=bibtexparser.bparser.BibTexParser())

  # write
  open(args['output'],'w').write(bibtexparser.dumps(bib))



"""GbibCheckAuthors
  Check for authors that are possibly stored in more than one way.

Usage:
  GbibCheckAuthors [options] <input>

Arguments:
  input                   Input BibTeX-file.

Options:
      --version           Show version.
  -h, --help              Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""

# ==================================================================================================

import os
import re
import sys
import bibtexparser
import docopt

from .. import __version__

# ======================================== REFORMAT AUTHORS ========================================

def reformatAuthor(text,sep=' '):

  # skip authors that cannot be split in first and last name
  if len(text.split(',')) <= 1:
    return text

  # define find/replace criteria
  match = [
    ( re.compile(r'(.*)(\(.*\))'                             , re.UNICODE) ,r'\1'          ),
    ( re.compile(r'([A-Za-z][\}]?)([\w0-9\{\}\'\"\\\.\^\{]*)', re.UNICODE) ,r'\1.'         ),
    ( re.compile(r'([A-Za-z\.][\-]?)([\ ]*)'                 , re.UNICODE) ,r'\1'          ),
    ( re.compile(r'([A-Za-z\.][\-]?)([A-Za-z])'              , re.UNICODE) ,r'\1'+sep+r'\2'),
  ]

  # extract first and last name
  last,first = text.split(',')

  # update extend all "." with a space, to distinguish initials
  first = first.replace('.','. ')

  # loop over over find/replace criteria
  for regex, sub in match:
    first = re.sub(regex, sub, first)

  # remove spaces from the beginning and end
  first = first.strip()

  # rejoin first and last name
  return last+', '+first.upper()

# ================================ CONVERT UNICODE SYMBOLS TO LATEX ================================

def replaceUnicode(text):

  # NB list not exhaustive, please extend!
  match = [
    ( u'ç', r'\c{c}'    ),
    ( u'è', r'\`{e}'    ),
    ( u'é', r'\'{e}'    ),
    ( u'ë', r'\"{e}'    ),
    ( u'ô', r'\^{o}'    ),
    ( u'ö', r'\"{o}'    ),
    ( u'ü', r'\"{y}'    ),
    ( u'g̃', r'\~{g}'    ),
    ( u'ñ', r'\~{n}'    ),
    ( u'İ', r'\.{I}'    ),
    ( u'à', r'\'{a}'    ),
    ( u'ă', r'\v{a}'    ),
    ( u'ř', r'\v{r}'    ),
    ( u'–',  '--'       ),
    ( u'—',  '--'       ),
    ( u'“',  '``'       ),
    ( u'”',  "''"       ),
    ( u'×', r'$\times$' ),
  ]

  # loop over over find/replace criteria
  for ex, sub in match:
    text = text.replace(ex, sub)

  return text

# ========================================== MAIN PROGRAM ==========================================

def main():

  # --------------------------------- parse command line arguments ---------------------------------

  # parse command-line options/arguments
  args = docopt.docopt(__doc__, version=__version__)

  # change keys to simplify implementation:
  # - remove leading "-" and "--" from options
  args = {re.sub(r'([\-]{1,2})(.*)',r'\2',key): args[key] for key in args}
  # - change "-" to "_" to facilitate direct use in print format
  args = {key.replace('-','_'): args[key] for key in args}
  # - remove "<...>"
  args = {re.sub(r'(<)(.*)(>)',r'\2',key): args[key] for key in args}

  # --------------------------------------- check arguments ----------------------------------------

  # check that the BibTeX file exists
  if not os.path.isfile(args['input']):
    print('"{input:s}" does not exist'.format(**args))
    sys.exit(1)

  # ---------------------------------------- parse bib-file ----------------------------------------

  # read
  # ----

  bib = bibtexparser.load(open(args['input'],'r'), parser=bibtexparser.bparser.BibTexParser())

  # check entries
  # -------------

  names = {}

  for entry in bib.entries:

    # set author separation
    sep = ''

    # fix author abbreviations
    for key in ['author','editor']:
      if key in entry:
        for name in entry[key].split(' and '):
          fmtName = replaceUnicode(reformatAuthor(name,sep))
          if fmtName not in names:
            names[fmtName]  = [name]
          if name not in names[fmtName]:
            names[fmtName] += [name]

  # filter unique
  names = {name:names[name] for name in names if len(names[name]) > 1}

  # output format
  n = max([len(name) for name in names])

  # print
  for name in sorted(names):
    print(('{0:%ds} : {1:s}'%(n)).format(name, '; '.join(names[name])))



"""GbibClean
  Clean a BibTeX database, stripping it from unnecessary fields, unifying the formatting of authors,
  and ensuring the proper special characters and math mode settings.

Usage:
  GbibClean [options] <input> <output>

Arguments:
  input                   Input BibTeX-file.
  output                  Output file, or path (the "input" file-name is copied).

Options:
      --author-sep=<str>  Character to separate authors' initials. [default: ]
      --ignore-case       Do not protect case of title.
      --ignore-math       Do not apply math-mode fix.
      --ignore-unicode    Do not apply unicode fix.
      --dot-space=<str>   Character separating abbreviation dots. [default: ]
      --no-title          Remove title from BibTeX file.
      --verbose-doi       Print entries from which the doi was automatically extracted.
      --verbose-arxivid   Print entries from which the arxivid was automatically extracted.
      --version           Show version.
  -h, --help              Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""

# ==================================================================================================

import os
import re
import sys
import bibtexparser
import docopt

from .. import __version__

# ==================================== RAISE COMMAND LINE ERROR ====================================

def Error(msg,exit_code=1):

  print(msg)

  sys.exit(exit_code)

# ===================================== RECURSIVE REPLACEMENT ======================================

def subr(regex, sub, text):

  # make substitutions, get the number of substitutions "n"
  text, n = re.subn(regex, sub, text)

  # continue substituting
  if n:
    return subr(regex, sub, text)

  return text

# ========================================== EXTRACT DOI ===========================================

def getDoi(*args):

  # define find/replace criteria
  match = [
    (re.compile(r'(.*)(https://doi.org/)([^\s]*)(.*)', re.IGNORECASE), r'\3'),
    (re.compile(r'(.*)(http://doi.org/)([^\s]*)(.*)' , re.IGNORECASE), r'\3'),
    (re.compile(r'(.*)(doi/abs/)([^\s]*)(.*)'        , re.IGNORECASE), r'\3'),
    (re.compile(r'(.*)(doi)([^\s]*)(.*)'             , re.IGNORECASE), r'\3'),
  ]

  # try to find matching entries (and strip them)
  for regex, sub in match:
    for arg in args:
      if re.match(regex, arg):
        num = re.sub(regex, sub, arg)
        if not re.match(r'(.*)([\s])(.*)',num): return num

  # quit function without a positive match
  return None

# ======================================== EXTRACT ARXIVID =========================================

def getArxiv(*args):

  # define find/replace criteria
  match = [
    (re.compile(r'(.*)(https://arxiv.org/abs/)([^\s]*)(.*)', re.IGNORECASE), r'\3'),
    (re.compile(r'(.*)(http://arxiv.org/abs/)([^\s]*)(.*)' , re.IGNORECASE), r'\3'),
    (re.compile(r'(.*)(arxiv)([^:]*)([:]?)(.*)'            , re.IGNORECASE), r'\5'),
    (re.compile(r'([0-9]*\.[0-9]*[v]?[0-9]*)'              , re.IGNORECASE), r'\1'),
  ]

  # try to find matching entries (and strip them)
  for regex, sub in match:
    for arg in args:
      if re.match(regex, arg):
        num = re.sub(regex, sub, arg)
        if not re.match(r'(.*)([\s])(.*)',num): return num

  # quit function without a positive match
  return None

# ======================================== REFORMAT AUTHORS ========================================

def reformatAuthor(text,sep=' '):

  # skip authors that cannot be split in first and last name
  if len(text.split(',')) <= 1:
    return text

  # define find/replace criteria
  match = [
    (re.compile(r'(.*)(\(.*\))'                               , re.UNICODE) ,r'\1'          ),
    (re.compile(r'([A-Za-z][\}]*)([\w0-9\{\}\`\'\"\\\.\^\{]*)', re.UNICODE) ,r'\1.'         ),
    (re.compile(r'([A-Za-z\.][\-]?)([\ ]*)'                   , re.UNICODE) ,r'\1'          ),
    (re.compile(r'([A-Za-z\.][\-]?)([A-Za-z])'                , re.UNICODE) ,r'\1'+sep+r'\2'),
  ]

  # extract first and last name
  last, first = text.split(',')

  # update extend all "." with a space, to distinguish initials
  first = first.replace('.', '. ')

  # loop over over find/replace criteria
  for regex, sub in match:
    first = re.sub(regex, sub, first)

  # remove spaces from the beginning and end
  first = first.strip()

  # rejoin first and last name
  return last + ', ' + first.upper()

# ========================================== PROTECT MATH ==========================================

def protectMath(text):

  # skip text without any math
  if len(text.split('{\$}')) < 3:
    return text

  # define find/replace criteria
  match = [
    (re.compile(r'(\{\\\$\})(.*)(\{\\\$\})'     ,re.UNICODE), r'$\2$'       ),
    (re.compile(r'(\$\$)(.*)(\$\$)'             ,re.UNICODE), r'$\2$'       ),
    (re.compile(r'(\$)(.*)(\{\\{\})(.*)(\$)'    ,re.UNICODE), r'\1\2{\4\5'  ),
    (re.compile(r'(\$)(.*)(\{\\}\})(.*)(\$)'    ,re.UNICODE), r'\1\2}\4\5'  ),
    (re.compile(r'(\$)(.*)(\{\\_\})(.*)(\$)'    ,re.UNICODE), r'\1\2_\4\5'  ),
    (re.compile(r'(\$)(.*)(\{\\^\})(.*)(\$)'    ,re.UNICODE), r'\1\2^\4\5'  ),
    (re.compile(r'(\$)(.*)(\\backslash)(.*)(\$)',re.UNICODE), r'\1\2\\\4\5' ),
  ]

  # loop over over find/replace criteria
  for regex, sub in match:
    text = subr(regex, sub, text)

  return text

# ================================ CONVERT UNICODE SYMBOLS TO LATEX ================================

def replaceUnicode(text):

  # NB list not exhaustive, please extend!
  match = [
    ( u'ç', r'\c{c}'    ),
    ( u'è', r'\`{e}'    ),
    ( u'é', r'\'{e}'    ),
    ( u'ë', r'\"{e}'    ),
    ( u'ô', r'\^{o}'    ),
    ( u'ö', r'\"{o}'    ),
    ( u'ü', r'\"{y}'    ),
    ( u'g̃', r'\~{g}'    ),
    ( u'ñ', r'\~{n}'    ),
    ( u'İ', r'\.{I}'    ),
    ( u'à', r'\'{a}'    ),
    ( u'ă', r'\v{a}'    ),
    ( u'ř', r'\v{r}'    ),
    ( u'–',  '--'       ),
    ( u'—',  '--'       ),
    ( u'“',  '``'       ),
    ( u'”',  "''"       ),
    ( u'×', r'$\times$' ),
  ]

  # loop over over find/replace criteria
  for ex, sub in match:
    text = text.replace(ex, sub)

  return text

# ========================================== MAIN PROGRAM ==========================================

def main():

  # --------------------------------- parse command line arguments ---------------------------------

  # parse command-line options/arguments
  args = docopt.docopt(__doc__, version=__version__)

  # change keys to simplify implementation:
  # - remove leading "-" and "--" from options
  args = {re.sub(r'([\-]{1,2})(.*)',r'\2',key): args[key] for key in args}
  # - change "-" to "_" to facilitate direct use in print format
  args = {key.replace('-','_'): args[key] for key in args}
  # - remove "<...>"
  args = {re.sub(r'(<)(.*)(>)',r'\2',key): args[key] for key in args}

  # --------------------------------------- check arguments ----------------------------------------

  # check that the BibTeX file exists
  if not os.path.isfile(args['input']):
    Error('"{input:s}" does not exist'.format(**args))

  # if "output" is an existing directory; convert to file with same name as the input file
  if os.path.isdir(args['output']):
    args['output'] = os.path.join(args['output'], os.path.split(args['input'])[-1])

  # ---------------------------------------- parse bib-file ----------------------------------------

  # read
  # ----

  bib = bibtexparser.load(open(args['input'],'r'), parser=bibtexparser.bparser.BibTexParser())

  # fix entries
  # -----------

  for entry in bib.entries:

    # fix known Mendeley bugs
    if 'journal' in entry:
      if entry['journal'] == 'EPL (Europhysics Lett.': entry['journal'] = 'EPL (Europhys. Lett.)'
      if entry['journal'] == 'Science (80-. ).'      : entry['journal'] = 'Science'

    # find doi
    if 'doi' not in entry:
      doi = getDoi(*[val for key,val in entry.items() if key not in ['arxivid', 'eprint']])
      if doi:
        entry['doi'] = doi
        if args['verbose_doi']: print(entry['ID'])

    # find arXiv-id
    if 'arxivid' not in entry:
      arxivid = getArxiv(*[val for key,val in entry.items() if key not in ['doi']])
      if arxivid:
        entry['arxivid'] = arxivid
        if args['verbose_arxivid']: print(entry['ID'])

    # get author separations
    sep = args['author_sep']

    # fix author abbreviations
    for key in ['author','editor']:
      if key in entry:
        entry[key] = ' and '.join([reformatAuthor(i,sep) for i in entry[key].split(' and ')])

    # remove title
    if args['no_title']:
      entry.pop('title',None)

    # protect math-mode
    if not args['ignore_math']:
      for key in ['title']:
        if key in entry:
          entry[key] = protectMath(entry[key])

    # convert unicode to LaTeX
    if not args['ignore_unicode']:
      for key in ['author','editor','title']:
        if key in entry:
          entry[key] = replaceUnicode(entry[key])

    # abbreviations: change symbol after "."
    for key in ['journal','author']:
      if key in entry:
        entry[key] = entry[key].replace('. ','.{dot_space:s} '.format(**args))

    # fix underscore problems
    # -
    if 'doi' in entry:
      entry['doi'] = entry['doi'].replace('_','\_')
      entry['doi'] = entry['doi'].replace('{\\\\_}','\_')
      entry['doi'] = entry['doi'].replace('{\\_}','\_')
      entry['doi'] = entry['doi'].replace('{\_}','\_')
    # -
    if 'url' in entry:
      entry['url'] = entry['url'].replace('{\_}','\_')
      entry['url'] = entry['url'].replace('{\\_}','\_')
      entry['url'] = entry['url'].replace('{~}','~')
      entry['url'] = entry['url'].replace(r'\&','&')
    # -
    if 'url' in entry:
      entry['url'] = subr(re.compile(r'({)([^}])(})',re.UNICODE), r'\2', entry['url'])

  # select entries
  # --------------

  for entry in bib.entries:

    # get entry type
    e = entry['ENTRYTYPE']

    # basic fields (common for all)
    fields = ['ID','ENTRYTYPE','author','title','year','doi','arxivid']

    # set fields to select (if present)
    if   e=='article'      : fields += ['journal','volume','number','pages']
    elif e=='unpublished'  : fields += []
    elif e=='inproceedings': fields += ['booktitle','editor','publisher','volume','number','pages']
    elif e=='book'         : fields += ['edition','editor','publisher','isbn']
    elif e=='inbook'       : fields += ['edition','editor','publisher','isbn']
    elif e=='incollection' : fields += ['booktitle','edition','publisher','isbn']
    elif e=='phdthesis'    : fields += ['school','isbn','url']
    elif e=='techreport'   : fields += ['institution','isbn','url']
    elif e=='misc'         : fields += ['pages','url']
    else: Error('Undefined entry '+entry)

    # ensure at least one digital reference
    if 'url' not in fields:
      if 'doi' not in entry and 'arxivid' not in entry:
        fields.append('url')

    # remove all fields that are not specified above
    rm = [key for key in entry if key not in fields]
    for key in rm:
      del entry[key]

    # remove link if doi or arxivid is available
    if 'url' in entry and ('doi' in entry or 'arxivid' in entry):
      del entry['url']

  # store
  # -----

  open(args['output'],'w').write(bibtexparser.dumps(bib))



"""GbibSelect
  Select BibTeX fields that are part of (a) TeX file(s).

Usage:
  GbibSelect [options] <tex>...

Options:
  -o, --output=<FILE>   Name of the output BibTeX file.
  -b, --bib=<FILE>      Name of the input BibTeX file.
      --version         Show version.
  -h, --help            Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""

# ==================================================================================================

import os
import re
import sys
import bibtexparser
import docopt

from .. import __version__

# ==================================== RAISE COMMAND LINE ERROR ====================================

def Error(msg,exit_code=1):

  print(msg)

  sys.exit(exit_code)

# ===================================== EXTRACT CITATION KEYS ======================================

def tex2cite(tex):

  # extract keys from "cite"
  def extract(s):
    try:
      return list(re.split('([pt])?(\[.*\]\[.*\])?(\{[a-zA-Z0-9\,\-\ ]*\})',s)[3][1:-1].split(','))
    except:
      print('Error in interpreting\n {0} ...').format(s[:100])

  # read all keys in "cite", "citet", "citep" commands
  cite = [extract(i) for i in tex.split(r'\cite')[1:]]
  cite = list(set([item for sublist in cite for item in sublist]))
  cite = [i.replace(' ','') for i in cite]

  return cite

# ========================================== MAIN PROGRAM ==========================================

def main():

  # --------------------------------- parse command line arguments ---------------------------------

  # parse command-line options/arguments
  args = docopt.docopt(__doc__, version=__version__)

  # change keys to simplify implementation:
  # - remove leading "-" and "--" from options
  args = {re.sub(r'([\-]{1,2})(.*)',r'\2',key): args[key] for key in args}
  # - change "-" to "_" to facilitate direct use in print format
  args = {key.replace('-','_'): args[key] for key in args}
  # - remove "<...>"
  args = {re.sub(r'(<)(.*)(>)',r'\2',key): args[key] for key in args}

  # --------------------------------------- check arguments ----------------------------------------

  # check that the TeX files exist
  for fname in args['tex']:
    if not os.path.isfile(fname):
      Error('"{0:s}" does not exist'.format(fname))

  # check that the BibTeX file exists
  if args['bib']:
    if not os.path.isfile(args['bib']):
      Error('"{0:s}" does not exist'.format(args['bib']))

  # if the file is an existing directory -> convert to file with same name as the input file
  if args['output']:
    if os.path.isdir(args['output']):
      args['output'] = os.path.join(args['output'], os.path.split(args['bib'])[-1])

  # -------------------------------------- get citation keys ---------------------------------------

  # initialize
  keys = []

  # read from TeX files
  for fname in args['tex']:
    keys += tex2cite(open(fname,'r').read())

  # sort citation keys
  keys = sorted(list(set(keys)))

  # optionally: print and quit
  if not args['bib'] and not args['output']:
    print('\n'.join(keys))

  # ------------------------------------- select from bib-file -------------------------------------

  # read BibTeX file
  bib = bibtexparser.load(open(args['bib'],'r'), parser=bibtexparser.bparser.BibTexParser())

  # get citation keys in the BibTeX files
  bibkeys = [entry['ID'] for entry in bib.entries]

  # find entries that are present in the TeX files, but missing in the BibTeX files
  missing = [key for key in keys if key not in bibkeys]

  # print missing entries
  if len(missing) > 0:
    Error('Not found: '+', '.join(missing))

  # select entries based on the keys found in the TeX file
  bib.entries = [entry for entry in bib.entries if entry['ID'] in keys]

  # convert to text
  bib = bibtexparser.dumps(bib)

  # write bib-file, or print to screen
  if args['output']: open(args['output'],'w').write(bib)
  else             : print(bib)



"""GbibLink
  Check if the "doi", "arxivid", and "url" in a BibTeX file exists. Note: this function has to
  actually browse to the sites, it is thus quite slow!

Usage:
  GbibLink [options] <bibfile>

Options:
  -n, --no-link   Print entries that do not have any of the listed links.
      --version   Show version.
  -h, --help      Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""

# ==================================================================================================

import os
import re
import sys
import bibtexparser
import docopt
import requests

from .. import __version__

# ==================================== RAISE COMMAND LINE ERROR ====================================

def Error(msg,exit_code=1):

  print(msg)

  sys.exit(exit_code)

# ====================================== CHECK IF URL EXISTS =======================================

def urlExists(url):

  request = requests.get(url)

  if request.status_code == 200: return True
  else                         : return False

# ========================================== MAIN PROGRAM ==========================================

def main():

  # --------------------------------- parse command line arguments ---------------------------------

  # parse command-line options/arguments
  args = docopt.docopt(__doc__, version=__version__)

  # change keys to simplify implementation:
  # - remove leading "-" and "--" from options
  args = {re.sub(r'([\-]{1,2})(.*)',r'\2',key): args[key] for key in args}
  # - change "-" to "_" to facilitate direct use in print format
  args = {key.replace('-','_'): args[key] for key in args}
  # - remove "<...>"
  args = {re.sub(r'(<)(.*)(>)',r'\2',key): args[key] for key in args}

  # --------------------------------------- check arguments ----------------------------------------

  # check that the BibTeX file exists
  if args['bibfile']:
    if not os.path.isfile(args['bibfile']):
      Error('"{0:s}" does not exist'.format(args['bibfile']))

  # ------------------------------------------ check file ------------------------------------------

  # read BibTeX file
  bib = bibtexparser.load(open(args['bibfile'],'r'), parser=bibtexparser.bparser.BibTexParser())

  # list with incorrect refs
  # - allocate
  doi = []
  # - check
  for entry in bib.entries:
    if 'doi' in entry:
      if not urlExists('http://dx.doi.org/'+entry['doi']):
        doi += [entry['ID']]

  # list with incorrect refs
  # - allocate
  arxivid = []
  # - check
  for entry in bib.entries:
    if 'arxivid' in entry:
      if not urlExists('http://arxiv.org/abs/'+entry['arxivid']):
        arxivid += [entry['ID']]

  # list with incorrect refs
  # - allocate
  url = []
  # - check
  for entry in bib.entries:
    if 'url' in entry:
      if not urlExists(entry['url']):
        url += [entry['ID']]

  # list entries without any link
  # - allocate
  nolink = []
  # - check
  if args['no_link']:
    for entry in bib.entries:
      if 'doi' not in entry and 'arxivid' not in entry and 'url' not in entry:
        nolink += [entry['ID']]

  # ---------------------------------------- print diagnosis -----------------------------------------

  diag = {}

  for key in doi    : diag[key] = diag.get(key,[]) + ['doi'    ]
  for key in arxivid: diag[key] = diag.get(key,[]) + ['arxivid']
  for key in url    : diag[key] = diag.get(key,[]) + ['url'    ]
  for key in nolink : diag[key] = diag.get(key,[]) + ['no-link']

  if len(diag) == 0:
    print('No errors')
    sys.exit(0)

  fmt = r'{key:%ds} : {value:s}' % max([len(key) for key in diag])

  for key in sorted(diag):
    print(fmt.format(key=key, value=', '.join(diag[key])))




"""GbibCheckKeys
  Check which entries do not have a citation key composed of "LastName + Year".

Usage:
  GbibCheckKeys [options] <input>

Arguments:
  input                   Input BibTeX-file.

Options:
  -s, --view-skipped      View skipped entries that cannot be composed in the standard way.
      --version           Show version.
  -h, --help              Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""

# ==================================================================================================

import os
import re
import sys
import bibtexparser
import docopt

from .. import __version__

# ==================================================================================================

def replaceUnicode(text):

    match = [
        (u'ç', 'c'),
        (u'è', 'e'),
        (u'é', 'e'),
        (u'ë', 'e'),
        (u'ô', 'o'),
        (u'ö', 'o'),
        (u'ü', 'y'),
        (u'g̃', 'g'),
        (u'ñ', 'n'),
        (u'İ', 'I'),
        (u'à', 'a'),
        (u'ă', 'a'),
        (u'ř', 'r'),
        (r'\c{c}', 'c'),
        (r'\`{e}', 'e'),
        (r'\'{e}', 'e'),
        (r'\"{e}', 'e'),
        (r'\^{o}', 'o'),
        (r'\"{o}', 'o'),
        (r'\"{y}', 'y'),
        (r'\~{g}', 'g'),
        (r'\~{n}', 'n'),
        (r'\.{I}', 'I'),
        (r'\'{a}', 'a'),
        (r'\v{a}', 'a'),
        (r'\v{r}', 'r'),
        (r'{\"{a}}', 'a'),
        (r'{\"{u}}', 'u'),
        (r'{\'{c}}', 'c'),
        (r'{\'{s}}', 's'),
        (r'{\^{i}}', 'i'),
        (r'{\v{s}}', 's'),
        (r'{\v{z}}', 'z'),
        (r'\aa', 'a'),
        (r'{\o}', 'o'),
        (r"'", ''),
        (r' ', ''),
    ]

    for ex, sub in match:
        text = text.replace(ex, sub)

    return text

# ==================================================================================================

def getPlainLastName(full_name):

    last = full_name.split(',')[0]

    last = replaceUnicode(last)
    last = last.replace('{', '')
    last = last.replace('}', '')
    last = last[0].upper() + last[1:]

    return last

# ========================================== MAIN PROGRAM ==========================================

def main():

    # parse command-line options/arguments
    # change keys to simplify implementation:
    # - remove leading "-" and "--" from options
    # - change "-" to "_" to facilitate direct use in print format
    # - remove "<...>"
    args = docopt.docopt(__doc__, version=__version__)
    args = {re.sub(r'([\-]{1,2})(.*)',r'\2',key): args[key] for key in args}
    args = {key.replace('-','_'): args[key] for key in args}
    args = {re.sub(r'(<)(.*)(>)',r'\2',key): args[key] for key in args}

    if not os.path.isfile(args['input']):
        print('"{input:s}" does not exist'.format(**args))
        sys.exit(1)

    alphabet = 'abcdefghijklmnopqrstuvwxyz'
    alphabet_list = list(alphabet)

    # get all keys and auto-generated equivalent

    bib = bibtexparser.load(open(args['input'],'r'), parser=bibtexparser.bparser.BibTexParser())

    keys = []

    for entry in bib.entries:

        old_key = entry['ID']
        new_key = entry['ID']
        generated = False

        if 'author' in entry and 'year' in entry:
            name = getPlainLastName(entry['author'].split(' and ')[0])
            year = entry['year']
            new_key = name + year
            generated = True

        if not generated:
            if 'editor' in entry and 'year' in entry:
                name = getPlainLastName(entry['editor'].split(' and ')[0])
                year = entry['year']
                new_key = name + year
                generated = True

        keys += [[old_key, new_key, generated]]

    keys = sorted(keys, key=lambda x: x[0])

    # handle duplicate keys by adding (a, b, c, ...); respect old

    ntimes = {new_key: 0 for _, new_key, _ in keys}

    for old_key, new_key, generated in keys:
        ntimes[new_key] += 1

    for idx, (old_key, new_key, generated) in enumerate(keys):
        if old_key != new_key:
            if ntimes[new_key] > 1:
                if re.match(r'[a-z]', old_key[-1]):
                    new_key += old_key[-1]
                    keys[idx][1] = new_key

    # handle duplicate keys by adding (a, b, c, ...)

    ntimes = {new_key: 0 for _, new_key, _ in keys}

    for old_key, new_key, generated in keys:
        ntimes[new_key] += 1

    for idx, (old_key, new_key, generated) in enumerate(keys):
        if not generated:
            continue
        if ntimes[new_key] > 1:
            ntimes[new_key] -= 1
            if not re.match(r'[a-z]', new_key[-1]):
                new_key += 'a'
            if new_key in ntimes:
                new_key = new_key[:-1] + alphabet_list[alphabet.find(new_key[-1]) + 1]
            keys[idx][1] = new_key
            ntimes[new_key] = 1

    # make sure that duplicate suffix (a, b, c, ...) is as low as possible

    ntimes = {new_key: 0 for _, new_key, _ in keys}

    for old_key, new_key, generated in keys:
        ntimes[new_key] += 1

    for idx, (old_key, new_key, generated) in enumerate(keys):

        if not generated:
            continue

        if not re.match(r'[a-z]', new_key[-1]):
            continue

        if new_key[:-1] not in ntimes:
            keys[idx][1] = new_key[:-1]
            ntimes[new_key[:-1]] = 1
            ntimes[new_key] -= 1
            continue

        if ntimes[new_key[:-1]] == 0:
            keys[idx][1] = new_key[:-1]
            ntimes[new_key[:-1]] += 1
            ntimes[new_key] -= 1
            continue

        for j in range(alphabet.find(new_key[:-1])):
            if new_key[:-1] + alphabet[j] not in ntimes:
                keys[idx][1] = new_key[:-1] + alphabet[j]
                ntimes[keys[idx][1]] = 1
                ntimes[new_key] -= 1
                break

            if ntimes[new_key[:-1] + alphabet[j]] == 0:
                keys[idx][1] = new_key[:-1] + alphabet[j]
                ntimes[keys[idx][1]] += 1
                ntimes[new_key] -= 1
                break

    # print respect

    n = max([len(old_key) for old_key, _, _ in keys])
    fmt = ('{0:%ds} : {1:s}'%(n))

    for old_key, new_key, generated in keys:
        if old_key != new_key:
            print(fmt.format(old_key, new_key))

    if args['view_skipped']:
        skipped = [new_key for old_key, new_key, generated in keys if not generated]
        if len(skipped) > 0:
            print('')
            print('Skipped keys:')
            print('\n'.join(skipped))





from setuptools import setup
from setuptools import find_packages
import re

filepath = 'GooseBib/__init__.py'
__version__ = re.findall(r'__version__ = \'(.*)\'', open(filepath).read())[0]

setup(
    name = 'GooseBib',
    version = __version__,
    license = 'MIT',
    author = 'Tom de Geus',
    author_email = 'tom@geus.me',
    description = 'Reformat BibTeX files',
    long_description = 'Reformat BibTeX files',
    keywords = 'LaTeX; BibTeX',
    url = 'https://github.com/tdegeus/GooseBib',
    packages = find_packages(),
    install_requires = ['docopt>=0.6.2', 'click>=4.0', 'bibtexparser>=1.0.0', 'requests>=2.0.0'],
    entry_points = {
        'console_scripts': [
            'GbibCheckAuthors = GooseBib.cli.GbibCheckAuthors:main',
            'GbibCheckLink = GooseBib.cli.GbibCheckLink:main',
            'GbibCheckKeys = GooseBib.cli.GbibCheckKeys:main',
            'GbibClean = GooseBib.cli.GbibClean:main',
            'GbibList = GooseBib.cli.GbibList:main',
            'GbibParse = GooseBib.cli.GbibParse:main',
            'GbibSelect = GooseBib.cli.GbibSelect:main',
            'GbibSelectAlias = GooseBib.cli.GbibSelectAlias:main']})



"""GbibList
  List a certain field for all entries.

Usage:
  GbibList [options] <input>

Arguments:
  input                   Input BibTeX-file.

Options:
  -j, --journal           List the journal field.
      --version           Show version.
  -h, --help              Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""

# ==================================================================================================

import os
import re
import sys
import bibtexparser
import docopt

from .. import __version__

# ========================================== MAIN PROGRAM ==========================================

def main():

  # --------------------------------- parse command line arguments ---------------------------------

  # parse command-line options/arguments
  args = docopt.docopt(__doc__, version=__version__)

  # change keys to simplify implementation:
  # - remove leading "-" and "--" from options
  args = {re.sub(r'([\-]{1,2})(.*)',r'\2',key): args[key] for key in args}
  # - change "-" to "_" to facilitate direct use in print format
  args = {key.replace('-','_'): args[key] for key in args}
  # - remove "<...>"
  args = {re.sub(r'(<)(.*)(>)',r'\2',key): args[key] for key in args}

  # --------------------------------------- check arguments ----------------------------------------

  # check that the BibTeX file exists
  if not os.path.isfile(args['input']):
    Error('"{input:s}" does not exist'.format(**args))

  # ---------------------------------------- parse bib-file ----------------------------------------

  # read
  # ----

  bib = bibtexparser.load(open(args['input'],'r'), parser=bibtexparser.bparser.BibTexParser())

  # list journals
  # -------------

  if args['journal']:

    out = []

    for entry in bib.entries:
      if 'journal' in entry:
        out += [entry['journal']]

    out = sorted(list(set(out)))

    print('\n'.join(out))



"""GbibSelect
  Select specific BibTeX fields and (optionally) choose the citation key. The selection is provided
  to the program using a json-file of the following format:

  ```json
  {
    "new_key" : "old_key",
    ...
  }
  ```

Usage:
  GbibSelect [options] <json>

Options:
  -o, --output=<FILE>   Name of the output BibTeX file.
  -b, --bib=<FILE>      Name of the input BibTeX file.
      --version         Show version.
  -h, --help            Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""

# ==================================================================================================

import os
import re
import sys
import bibtexparser
import docopt
import json

from .. import __version__

# ==================================== RAISE COMMAND LINE ERROR ====================================

def Error(msg,exit_code=1):

  print(msg)

  sys.exit(exit_code)

# ========================================== MAIN PROGRAM ==========================================

# ---------------------------------- parse command line arguments ----------------------------------

# parse command-line options/arguments
args = docopt.docopt(__doc__, version=__version__)

# change keys to simplify implementation:
# - remove leading "-" and "--" from options
args = {re.sub(r'([\-]{1,2})(.*)',r'\2',key): args[key] for key in args}
# - change "-" to "_" to facilitate direct use in print format
args = {key.replace('-','_'): args[key] for key in args}
# - remove "<...>"
args = {re.sub(r'(<)(.*)(>)',r'\2',key): args[key] for key in args}

# ---------------------------------------- check arguments -----------------------------------------

# check that the TeX files exist
if not os.path.isfile(args['json']):
  Error('"{0:s}" does not exist'.format(args['json']))

# check that the BibTeX file exists
if args['bib']:
  if not os.path.isfile(args['bib']):
    Error('"{0:s}" does not exist'.format(args['bib']))

# if the file is an existing directory -> convert to file with same name as the input file
if args['output']:
  if os.path.isdir(args['output']):
    args['output'] = os.path.join(args['output'], os.path.split(args['bib'])[-1])

# -------------------------------------- select from bib-file --------------------------------------

# data
with open(args['json']) as json_file:
  data = json.load(json_file)

# inverse data
data_inv = {data[key]: key for key in data}

# extract lists
keys_new = [key       for key in sorted(data)]
keys_old = [data[key] for key in sorted(data)]

# read BibTeX file
bib = bibtexparser.load(open(args['bib'],'r'), parser=bibtexparser.bparser.BibTexParser())

# get citation keys in the BibTeX files
bibkeys = [entry['ID'] for entry in bib.entries]

# find entries that are present in the TeX files, but missing in the BibTeX files
missing = [key for key in keys_old if key not in bibkeys]

# print missing entries
if len(missing) > 0:
  Error('Not found: '+', '.join(missing))

# select entries based on the keys found in the TeX file
bib.entries = [entry for entry in bib.entries if entry['ID'] in keys_old]

# modify citation keys
for entry in bib.entries:
  entry['ID'] = data_inv[entry['ID']]

# convert to text
bib = bibtexparser.dumps(bib)

# write bib-file, or print to screen
if args['output']: open(args['output'],'w').write(bib)
else             : print(bib)
